@@ -39,9 +39,10 @@ from typing import TYPE_CHECKING
 
 from libqtile import bar, configurable, confreader
 from libqtile.command import interface
-from libqtile.command.base import CommandError, CommandObject
+from libqtile.command.base import CommandError, CommandObject, expose_command
 from libqtile.lazy import LazyCall
 from libqtile.log_utils import logger
+from libqtile.utils import ColorType, create_task
 
 if TYPE_CHECKING:
     from typing import Any
@@ -108,7 +109,7 @@ class _Widget(CommandObject, configurable.Configurable):
     you need access to the qtile object, it needs to be imported into your code.
 
     ``lazy`` functions can also be passed as callback functions and can be used in
-    the same was as keybindings.
+    the same way as keybindings.
 
     For example:
 
@@ -117,7 +118,7 @@ class _Widget(CommandObject, configurable.Configurable):
         from libqtile import qtile
 
         def open_calendar():
-            qtile.cmd_spawn('gsimplecal next_month')
+            qtile.spawn('gsimplecal next_month')
 
         clock = widget.Clock(
             mouse_callbacks={
@@ -138,14 +139,14 @@ class _Widget(CommandObject, configurable.Configurable):
 
     offsetx: int = 0
     offsety: int = 0
-    defaults = [
+    defaults: list[tuple[str, Any, str]] = [
         ("background", None, "Widget background color"),
         (
             "mouse_callbacks",
             {},
             "Dict of mouse button press callback functions. Accepts functions and ``lazy`` calls.",
         ),
-    ]  # type: list[tuple[str, Any, str]]
+    ]
 
     def __init__(self, length, **config):
         """
@@ -169,8 +170,9 @@ class _Widget(CommandObject, configurable.Configurable):
             raise confreader.ConfigError("Widget width must be an int")
 
         self.configured = False
-        self._futures: list[asyncio.TimerHandle] = []
+        self._futures: list[asyncio.Handle] = []
         self._mirrors: set[_Widget] = set()
+        self.finalized = False
 
     @property
     def length(self):
@@ -186,12 +188,12 @@ class _Widget(CommandObject, configurable.Configurable):
     def width(self):
         if self.bar.horizontal:
             return self.length
-        return self.bar.size - (self.bar.border_width[1] + self.bar.border_width[3])
+        return self.bar.width
 
     @property
     def height(self):
         if self.bar.horizontal:
-            return self.bar.size - (self.bar.border_width[0] + self.bar.border_width[2])
+            return self.bar.height
         return self.length
 
     @property
@@ -215,7 +217,6 @@ class _Widget(CommandObject, configurable.Configurable):
     def timer_setup(self):
         """This is called exactly once, after the widget has been configured
         and timers are available to be set up."""
-        pass
 
     def _configure(self, qtile, bar):
         self._test_orientation_compatibility(bar.horizontal)
@@ -223,9 +224,18 @@ class _Widget(CommandObject, configurable.Configurable):
         self.qtile = qtile
         self.bar = bar
         self.drawer = bar.window.create_drawer(self.bar.width, self.bar.height)
+
+        # Clear this flag as widget may be restarted (e.g. if screen removed and re-added)
+        self.finalized = False
+
+        # Timers are added to futures list so they can be cancelled if the `finalize` method is
+        # called before the timers have fired.
         if not self.configured:
-            self.qtile.call_soon(self.timer_setup)
-            self.qtile.call_soon(asyncio.create_task, self._config_async())
+            timer = self.qtile.call_soon(self.timer_setup)
+            async_timer = self.qtile.call_soon(asyncio.create_task, self._config_async())
+
+            # Add these to our list of futures so they can be cancelled.
+            self._futures.extend([timer, async_timer])
 
     async def _config_async(self):
         """
@@ -236,7 +246,6 @@ class _Widget(CommandObject, configurable.Configurable):
         wish to initialise the relevant code (e.g. connections to dbus
         using dbus_next) here.
         """
-        pass
 
     def finalize(self):
         for future in self._futures:
@@ -244,12 +253,19 @@ class _Widget(CommandObject, configurable.Configurable):
         if hasattr(self, "layout") and self.layout:
             self.layout.finalize()
         self.drawer.finalize()
+        self.finalized = True
+
+        # Reset configuration status so the widget can be reconfigured
+        # e.g. when screen is re-added
+        self.configured = False
 
     def clear(self):
         self.drawer.set_source_rgb(self.bar.background)
         self.drawer.fillrect(self.offsetx, self.offsety, self.width, self.height)
 
+    @expose_command()
     def info(self):
+        """Info for this object."""
         return dict(
             name=self.name,
             offset=self.offset,
@@ -264,13 +280,13 @@ class _Widget(CommandObject, configurable.Configurable):
         self.mouse_callbacks = defaults
 
     def button_press(self, x, y, button):
-        name = "Button{0}".format(button)
+        name = f"Button{button}"
         if name in self.mouse_callbacks:
             cmd = self.mouse_callbacks[name]
             if isinstance(cmd, LazyCall):
                 if cmd.check(self.qtile):
                     status, val = self.qtile.server.call(
-                        (cmd.selectors, cmd.name, cmd.args, cmd.kwargs)
+                        (cmd.selectors, cmd.name, cmd.args, cmd.kwargs, False)
                     )
                     if status in (interface.ERROR, interface.EXCEPTION):
                         logger.error("Mouse callback command error %s: %s", cmd.name, val)
@@ -286,7 +302,7 @@ class _Widget(CommandObject, configurable.Configurable):
         """
         w = q.widgets_map.get(name)
         if not w:
-            raise CommandError("No such widget: %s" % name)
+            raise CommandError(f"No such widget: {name}")
         return w
 
     def _items(self, name: str) -> ItemT:
@@ -302,12 +318,6 @@ class _Widget(CommandObject, configurable.Configurable):
         elif name == "screen":
             return self.bar.screen
 
-    def cmd_info(self):
-        """
-        Info for this object.
-        """
-        return self.info()
-
     def draw(self):
         """
         Method that draws the widget. You may call this explicitly to
@@ -330,6 +340,10 @@ class _Widget(CommandObject, configurable.Configurable):
         """
         This method calls ``.call_later`` with given arguments.
         """
+        # Don't add timers for finalised widgets
+        if self.finalized:
+            return
+
         future = self.qtile.call_later(seconds, self._wrapper, method, *method_args)
 
         self._futures.append(future)
@@ -345,16 +359,38 @@ class _Widget(CommandObject, configurable.Configurable):
 
     def _remove_dead_timers(self):
         """Remove completed and cancelled timers from the list."""
+
+        def is_ready(timer):
+            return timer in self.qtile._eventloop._ready
+
         self._futures = [
             timer
             for timer in self._futures
-            if not (timer.cancelled() or timer.when() < self.qtile._eventloop.time())
+            # Filter out certain handles...
+            if not (
+                timer.cancelled()
+                # Once a scheduled timer is ready to be run its _scheduled flag is set to False
+                # and it's added to the loop's `_ready` queue
+                or (
+                    isinstance(timer, asyncio.TimerHandle)
+                    and not timer._scheduled
+                    and not is_ready(timer)
+                )
+                # Callbacks scheduled via `call_soon` are put into the loop's `_ready` queue
+                # and are removed once they've been executed
+                or (isinstance(timer, asyncio.Handle) and not is_ready(timer))
+            )
         ]
 
     def _wrapper(self, method, *method_args):
         self._remove_dead_timers()
         try:
-            method(*method_args)
+            if asyncio.iscoroutinefunction(method):
+                create_task(method(*method_args))
+            elif asyncio.iscoroutine(method):
+                create_task(method)
+            else:
+                method(*method_args)
         except:  # noqa: E722
             logger.exception("got exception from widget timer")
 
@@ -362,7 +398,7 @@ class _Widget(CommandObject, configurable.Configurable):
         return Mirror(self, background=self.background)
 
     def clone(self):
-        return copy.copy(self)
+        return copy.deepcopy(self)
 
     def mouse_enter(self, x, y):
         pass
@@ -408,9 +444,6 @@ class _Widget(CommandObject, configurable.Configurable):
                 del self._old_draw
 
 
-UNSPECIFIED = bar.Obj("UNSPECIFIED")
-
-
 class _TextBox(_Widget):
     """
     Base class for widgets that are just boxes containing text.
@@ -427,10 +460,9 @@ class _TextBox(_Widget):
         (
             "fmt",
             "{}",
-            "To format the string returned by the widget. For example, if the clock widget \
-             returns '08:46' we can do fmt='time {}' do print 'time 08:46' on the widget. \
-             To format the individual strings like hour and minutes use the format paramater \
-             of the widget (if it has one)",
+            "Format to apply to the string returned by the widget. Main purpose: applying markup. "
+            "For a widget that returns ``foo``, using ``fmt='<i>{}</i>'`` would give you ``<i>foo</i>``. "
+            "To control what the widget outputs in the first place, use the ``format`` paramater of the widget (if it has one).",
         ),
         ("max_chars", 0, "Maximum number of characters to display in widget."),
         (
@@ -456,6 +488,12 @@ class _TextBox(_Widget):
             "Whether text should scroll completely away (True) or stop when the end of the text is shown (False)",
         ),
         ("scroll_hide", False, "Whether the widget should hide when scrolling has finished"),
+        (
+            "scroll_fixed_width",
+            False,
+            "When ``scroll=True`` the ``width`` parameter is a maximum width and, when text is shorter than this, the widget will resize. "
+            "Setting ``scroll_fixed_width=True`` will force the widget to have a fixed width, regardless of the size of the text.",
+        ),
     ]  # type: list[tuple[str, Any, str]]
 
     def __init__(self, text=" ", width=bar.CALCULATED, **config):
@@ -556,7 +594,11 @@ class _TextBox(_Widget):
             self._is_scrolling = True
             self._should_scroll = True
         else:
-            self.length_type = bar.CALCULATED
+            if self.scroll_fixed_width:
+                self.length_type = bar.STATIC
+                self.length = self._scroll_width
+            else:
+                self.length_type = bar.CALCULATED
             self._should_scroll = False
 
     def calculate_length(self):
@@ -669,19 +711,26 @@ class _TextBox(_Widget):
     def hide_scroll(self):
         self.update("")
 
-    def cmd_set_font(self, font=UNSPECIFIED, fontsize=UNSPECIFIED, fontshadow=UNSPECIFIED):
+    @expose_command()
+    def set_font(
+        self,
+        font: str | None = None,
+        fontsize: int = 0,
+        fontshadow: ColorType = "",
+    ):
         """
         Change the font used by this widget. If font is None, the current
         font is used.
         """
-        if font is not UNSPECIFIED:
+        if font is not None:
             self.font = font
-        if fontsize is not UNSPECIFIED:
+        if fontsize != 0:
             self.fontsize = fontsize
-        if fontshadow is not UNSPECIFIED:
+        if fontshadow != "":
             self.fontshadow = fontshadow
         self.bar.draw()
 
+    @expose_command()
     def info(self):
         d = _Widget.info(self)
         d["foreground"] = self.foreground
@@ -689,6 +738,14 @@ class _TextBox(_Widget):
         return d
 
     def update(self, text):
+        """Update the widget text."""
+        # Don't try to update text in dead layouts
+        # This is mainly required for ThreadPoolText based widgets as the
+        # polling function cannot be cancelled and so may be called after the widget
+        # is finalised.
+        if not self.can_draw():
+            return
+
         if self.text == text:
             return
         if text is None:
@@ -717,8 +774,7 @@ class InLoopPollText(_TextBox):
         (
             "update_interval",
             600,
-            "Update interval in seconds, if none, the "
-            "widget updates whenever the event loop is idle.",
+            "Update interval in seconds, if none, the widget updates only once.",
         ),
     ]  # type: list[tuple[str, Any, str]]
 
@@ -773,7 +829,7 @@ class ThreadPoolText(_TextBox):
         (
             "update_interval",
             600,
-            "Update interval in seconds, if none, the " "widget updates whenever it's done.",
+            "Update interval in seconds, if none, the widget updates only once.",
         ),
     ]  # type: list[tuple[str, Any, str]]
 
@@ -795,13 +851,11 @@ class ThreadPoolText(_TextBox):
 
                     if self.update_interval is not None:
                         self.timeout_add(self.update_interval, self.timer_setup)
-                    else:
-                        self.timer_setup()
 
                 except Exception:
-                    logger.exception("Failed to reschedule.")
+                    logger.exception("Failed to reschedule timer for %s.", self.name)
             else:
-                logger.warning("poll() returned None, not rescheduling")
+                logger.warning("%s's poll() returned None, not rescheduling", self.name)
 
         self.future = self.qtile.run_in_executor(self.poll)
         self.future.add_done_callback(on_done)
@@ -809,7 +863,8 @@ class ThreadPoolText(_TextBox):
     def poll(self):
         pass
 
-    def cmd_force_update(self):
+    @expose_command()
+    def force_update(self):
         """Immediately poll the widget. Existing timers are unaffected."""
         self.update(self.poll())
 
@@ -881,8 +936,9 @@ class Mirror(_Widget):
         _Widget.__init__(self, reflection.length, **config)
         self.reflects = reflection
         self._length = 0
-        if self.reflects.length_type == bar.STRETCH:
-            self.length_type = bar.STRETCH
+        self.length_type = self.reflects.length_type
+        if self.length_type is bar.STATIC:
+            self._length = self.reflects._length
 
     def _configure(self, qtile, bar):
         _Widget._configure(self, qtile, bar)
@@ -891,6 +947,9 @@ class Mirror(_Widget):
         # there's no reflection, the mirror matches its parent bar.
         self.drawer.clear(self.background or self.bar.background)
 
+    def calculate_length(self):
+        return self.reflects.calculate_length()
+
     @property
     def length(self):
         if self.length_type != bar.STRETCH:
@@ -902,13 +961,11 @@ class Mirror(_Widget):
         self._length = value
 
     def draw(self):
-        if self.length_type != bar.STRETCH and self._length != self.reflects.length:
-            self._length = self.length
-            self.bar.draw()
-        else:
-            self.drawer.clear(self.reflects.background or self.bar.background)
-            self.reflects.drawer.paint_to(self.drawer)
-            self.drawer.draw(offsetx=self.offset, offsety=self.offsety, width=self.width)
+        if self.length <= 0:
+            return
+        self.drawer.clear_rect()
+        self.reflects.drawer.paint_to(self.drawer)
+        self.drawer.draw(offsetx=self.offset, offsety=self.offsety, width=self.width)
 
     def button_press(self, x, y, button):
         self.reflects.button_press(x, y, button)
